use std::net::{Ipv4Addr};

use aya::{Bpf, maps::HashMap};
use anyhow::{anyhow, Context};
use aya::programs::{Xdp, XdpFlags};
use aya_log::BpfLogger;
use clap::Parser;
use log::{info, warn};
use tokio::signal;
use csv::ReaderBuilder;
use serde::Deserialize;

#[derive(Debug, Parser)]
struct Opt {
    #[clap(short, long, default_value = "lo")]
    iface: String,
    #[clap(short, long, default_value = "icmp")]
    scan_type: String,
    #[clap(short, long)]
    csv: Option<String>,
}

#[derive(Debug, Deserialize, Eq, PartialEq)]
struct CsvRow {
    saddr: Ipv4Addr,
}

#[tokio::main]
async fn main() -> Result<(), anyhow::Error> {
    let opt = Opt::parse();

    env_logger::init();

    #[cfg(debug_assertions)]
    let bpefl_dir =  "target/bpfel-unknown-none/debug";

    #[cfg(not(debug_assertions))]
    let bpefl_dir =  "target/bpfel-unknown-none/release";

    let xdp_name = opt.scan_type;

    let mut bpf = Bpf::load_file(format!("{}/{}",
        bpefl_dir, xdp_name
    ))?;
    // TODO Proper pathing

    if let Err(e) = BpfLogger::init(&mut bpf) {
        // This can happen if you remove all log statements from your eBPF program.
        warn!("failed to initialize eBPF logger: {}", e);
    }

    // Obtain and load the XDP program called "responder" defined in the XDP file loaded above
    let program: &mut Xdp = bpf.program_mut("responder").unwrap().try_into()?;
    program.load()?;
    program.attach(&opt.iface, XdpFlags::default())
           .context(format!(
               "failed to attach the {} XDP program with default flags - try changing XdpFlags::default() to XdpFlags::SKB_MODE",
               xdp_name
           ))?;

    info!("Loaded {} XDP program", xdp_name);

    let mut filter_map: HashMap<_, u32, u8> = HashMap::try_from(bpf.map_mut("FILTER_MAP").ok_or(anyhow!("Could not construct mutable FILTER_MAP"))?)?;

    if let Some(csv_path) = opt.csv {
        info!("Installing filter rules from {}", csv_path);
        let mut reader = ReaderBuilder::new()
            .has_headers(true)
            .from_path(csv_path)?;
        for record in reader.deserialize() {
            let row: CsvRow = record?;
            filter_map.insert(u32::from(row.saddr), 1u8, 0)?;
        }
    };

    info!("press ctrl-c to exit");
    signal::ctrl_c().await?;
    info!("Exiting...");

    Ok(())
}
